//  SpinningEarth-D3D.cpp
//
//       YOU DO NOT NEED TO READ THIS FILE TO UNDERSTAND
//             COMPUTER GRAPHICS IN CURVED SPACES.
//
//  CurvedSpaces-Simple includes the much simpler SceneryObjectCube
//  and SceneryObjectOctahedron, which illustrate the basic principles.

#include "SpinningEarth-D3D.hpp"
#include <assert.h>

#define RADIANS_PER_SECOND  (D3DX_PI/4)
#define EARTH_RADIUS        0.05f

//  The standard D3DXCreateSphere() is easy to use,
//  but its rectangular triangulation scheme, aligned along
//  latitude and longitude lines, is woefully inefficient,
//  especially near the poles.  Use an efficient icosahedral
//  triangulation instead, subdivided as necessary.


CSpinningEarth::CSpinningEarth(void)
{
    int i;

    itsRotationAngle = 0.0;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        itsVertices[i]      = NULL;
        itsFaces[i]         = NULL;
        itsNumVertices[i]   = 0;
        itsNumFaces[i]      = 0;
    }

    InitIcosahedron();

    for (i = 0; i < REFINEMENT_LEVEL; i++)
        SubdivideTriangulation(i);
}

CSpinningEarth::~CSpinningEarth(void)
{
    int i;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        SAFE_DELETE_(itsVertices[i])
        SAFE_DELETE_(itsFaces[i])
    }

}


void CSpinningEarth::Update(double anElapsedTime)
{
    itsRotationAngle += RADIANS_PER_SECOND * anElapsedTime;

    if (itsRotationAngle > 2*D3DX_PI)
        itsRotationAngle -= 2*D3DX_PI;
}


void CSpinningEarth::Render(
    LPDIRECT3DDEVICE7       aDevice,
    D3DXMATRIX              *aPosition,
    D3DXMATRIX              *aCameraPosition,
    LPDIRECTDRAWSURFACE7    aTexture,
    DWORD                   aColor,
    int                     aFogRange,
    DWORD                   (*aFogFunction)(float *, float *, int, float *))
{
    D3DXMATRIX          theRotation,
                        theQuarterTurn,
                        theAttitude,
                        theWorldMatrix;
    DWORD               theFogFactor;
    float               theDistance;
    int                 theRefinementLevel,
                        i;

    //  For simplicity use a single fog factor for the whole Earth.
    theFogFactor = (*aFogFunction)( aCameraPosition->m[3],
                                    aPosition->m[3],
                                    aFogRange,
                                    &theDistance);

    try
    {
        D3DXMatrixRotationZ(&theRotation, (float) itsRotationAngle);
        D3DXMatrixRotationX(&theQuarterTurn, D3DX_PI/2);
        D3DXMatrixMultiply(&theAttitude, &theRotation, &theQuarterTurn);
        D3DXMatrixMultiply(&theWorldMatrix, &theAttitude, aPosition);
        D3DTRY(aDevice->SetTransform(D3DTRANSFORMSTATE_WORLD, theWorldMatrix))

        D3DTRY(aDevice->SetTexture(0, aTexture))
        D3DTRY(aDevice->SetRenderState(D3DRENDERSTATE_WRAP0, D3DWRAPCOORD_0))

        //  Pick an appropriate level of detail at which
        //  to render this image of the spinning Earth.
        assert(REFINEMENT_LEVEL >= 2);
        if (theDistance < 1.0)
            theRefinementLevel = 2;
        else
        if (theDistance < 1.58) //  just above pi/2
            theRefinementLevel = 1;
        else
            theRefinementLevel = 0;

        for (i = 0; i < itsNumVertices[theRefinementLevel]; i++)
        {
            itsVertices[theRefinementLevel][i].specular   = theFogFactor;
            itsVertices[theRefinementLevel][i].color      = aColor;
        }

        D3DTRY(aDevice->DrawIndexedPrimitive(
                    D3DPT_TRIANGLELIST,
                    D3DFVF_LVERTEX,
                    itsVertices[theRefinementLevel],
                    itsNumVertices[theRefinementLevel],
                    itsFaces[theRefinementLevel],
                    3*itsNumFaces[theRefinementLevel],
                    0))
    }
    catch (CException *anException)
    {
        MessageBox(NULL, anException->Message(), "Rendering Error in SpinningEarth", MB_OK);
        SAFE_DELETE(anException)
    }
    catch (...)
    {
        MessageBox(NULL, "Unknown exception.", "Rendering Error in SpinningEarth", MB_OK);
    }
}


//  The remaining methods implement the icosahedral triangulation
//  of the sphere.  Feel free to ignore them.

void CSpinningEarth::InitIcosahedron(void)
{
    float   v[12][3],
            d,
            s,
            t;
    int     i,
            ii;
    WORD    *f;

    itsNumVertices[0]   = 12;
    itsNumFaces[0]      = 20;
    itsVertices[0]      = new D3DLVERTEX[12];
    itsFaces[0]         = new WORD[3*20];

    assert(itsVertices[0] != NULL);
    assert(itsFaces[0]    != NULL);

    d = (float) acos
    (
        (cos(2*D3DX_PI/5)*cos(2*D3DX_PI/5) + cos(2*D3DX_PI/5))
      / (sin(2*D3DX_PI/5)*sin(2*D3DX_PI/5))
    );

    v[ 0][0] =  0.0f;
    v[ 0][1] =  0.0f;
    v[ 0][2] =  1.0f;

    v[11][0] =  0.0f;
    v[11][1] =  0.0f;
    v[11][2] = -1.0f;

    for (i = 0; i < 5; i++)
    {
        s = i * 2*D3DX_PI/5;
        t = s + D3DX_PI/5;

        v[1+i][0] = (float)(  cos(s) * sin(d)   );
        v[1+i][1] = (float)(  sin(s) * sin(d)   );
        v[1+i][2] = (float)( +cos(d)            );

        v[6+i][0] = (float)(  cos(t) * sin(d)   );
        v[6+i][1] = (float)(  sin(t) * sin(d)   );
        v[6+i][2] = (float)( -cos(d)            );
    }

    for (i = 0; i < 12; i++)
    {
        itsVertices[0][i].x  = v[i][0] * EARTH_RADIUS;
        itsVertices[0][i].y  = v[i][1] * EARTH_RADIUS;
        itsVertices[0][i].z  = v[i][2] * EARTH_RADIUS;

//      itsVertices[0][i].nx = v[i][0];
//      itsVertices[0][i].ny = v[i][1];
//      itsVertices[0][i].nz = v[i][2];

        itsVertices[0][i].color     = 0xFFFFFFFF;
        itsVertices[0][i].specular  = 0xFFFFFFFF;

        ComputeTextureCoordinates(v[i][0], v[i][1], v[i][2], &itsVertices[0][i]);
    }

    for (i = 0; i < 5; i++)
    {
        ii = (i + 1) % 5;

        f = itsFaces[0] + i*12;

        f[ 0] = 0;
        f[ 1] = i+1;
        f[ 2] = ii+1;

        f[ 3] = i+6;
        f[ 4] = ii+1;
        f[ 5] = i+1;

        f[ 6] = i+6;
        f[ 7] = ii+6;
        f[ 8] = ii+1;

        f[ 9] = 11;
        f[10] = ii+6;
        f[11] = i+6;
    }
}

void CSpinningEarth::ComputeTextureCoordinates(
    float       x,
    float       y,
    float       z,
    D3DLVERTEX  *aVertex)
{
    //  Technical note:  The north pole maps to texture
    //  coordinates (0.5, 1.0).  Thus half the arctic region
    //  is lost, and there's a texture discontinuity along
    //  the edge between two polar triangles.  This problem
    //  could be corrected by compressing a strip along
    //  the top edge of the bitmap from a rectangle to a
    //  triangle, but it's probably not worth the effort
    //  because (1) the discontinuity is small for a second order
    //  subdivision, and (2) the polar bitmap looks a bit
    //  pentagonal in any case.  Nevertheless, if in the future
    //  I want to fix the discontinuity, it can be done using
    //  DirectDrawSuface7::Blt().

    //  In theory atan2(0,0) returns 0, but who knows whether
    //  some future version of the math library might make
    //  trouble for us.
    if (x != 0.0f || y != 0.0f)
        aVertex->tu = (float) (atan2(y, x)/(2*D3DX_PI) + 0.5);
    else
        aVertex->tu = 0.5f;

    aVertex->tv = (float) (asin(z)/D3DX_PI + 0.5);
    
    //  For (unnecessary?) insurance against rounding errors...
    if (aVertex->tu < 0.0f)
        aVertex->tu = 0.0f;
    if (aVertex->tu > 1.0f)
        aVertex->tu = 1.0f;
    if (aVertex->tv < 0.0f)
        aVertex->tv = 0.0f;
    if (aVertex->tv > 1.0f)
        aVertex->tv = 1.0f;
}

void CSpinningEarth::SubdivideTriangulation(int aLevel)
{
    //  Subdivide the triangulation, replacing
    //  each old triangle with four new ones.
    //
    //    /\
    //   /__\
    //  /_\/_\
    //
    //  Write the subdivision to aLevel+1.

    WORD        u,
                v,
                *f,
                ff[3];
    D3DVECTOR   normal;
    int         *theTable,
                theVertexCount,
                i,
                j;

    assert( aLevel >=0
         && aLevel < REFINEMENT_LEVEL);

    itsNumVertices[aLevel+1]    = itsNumVertices[aLevel] + 3*itsNumFaces[aLevel]/2;
    itsNumFaces[aLevel+1]       = 4 * itsNumFaces[aLevel];
    itsVertices[aLevel+1]       = new D3DLVERTEX[itsNumVertices[aLevel+1]];
    itsFaces[aLevel+1]          = new WORD[3*itsNumFaces[aLevel+1]];

    assert(itsVertices[aLevel+1] != NULL);
    assert(itsFaces[aLevel+1]    != NULL);

    //  Copy the vertices from the previous level.
    for (i = 0; i < itsNumVertices[aLevel]; i++)
        itsVertices[aLevel+1][i] = itsVertices[aLevel][i];
    theVertexCount = itsNumVertices[aLevel];

    //  Create one new vertex on each edge.
    //  Make a table to index them, so the two triangles sharing
    //  an edge can share the same vertex.  (The memory for the
    //  table grows as the square of the number of old vertices.
    //  For modest triangulations this won't be a problem.
    //  For large triangulations a fancier algorithm could be used.)

    //  In spirit theTable is a 2-dimensional array.
    //  But such an array is a hassle to allocate in C/C++,
    //  so instead we just replace theTable[i][j] with
    //  theTable[i*itsNumVertices[aLevel] + j].
    theTable = new int[itsNumVertices[aLevel] * itsNumVertices[aLevel]];
    assert(theTable != NULL);

    for (i = 0; i < itsNumVertices[aLevel]; i++)
        for (j = 0; j < itsNumVertices[aLevel]; j++)
            theTable[i*itsNumVertices[aLevel] + j] = -1;

    for (i = 0; i < itsNumFaces[aLevel]; i++)
        for (j = 0; j < 3; j++)
        {
            u = itsFaces[aLevel][3*i +   j    ];
            v = itsFaces[aLevel][3*i + (j+1)%3];

            if (theTable[u*itsNumVertices[aLevel] + v] == -1)
            {
                assert(theVertexCount < itsNumVertices[aLevel+1]);

                normal.x = (itsVertices[aLevel][u].x + itsVertices[aLevel][v].x)/2;
                normal.y = (itsVertices[aLevel][u].y + itsVertices[aLevel][v].y)/2;
                normal.z = (itsVertices[aLevel][u].z + itsVertices[aLevel][v].z)/2;

                normal = Normalize(normal);

                itsVertices[aLevel+1][theVertexCount].x  = normal.x * EARTH_RADIUS;
                itsVertices[aLevel+1][theVertexCount].y  = normal.y * EARTH_RADIUS;
                itsVertices[aLevel+1][theVertexCount].z  = normal.z * EARTH_RADIUS;

//              itsVertices[aLevel+1][theVertexCount].nx = normal.x;
//              itsVertices[aLevel+1][theVertexCount].ny = normal.y;
//              itsVertices[aLevel+1][theVertexCount].nz = normal.z;

                itsVertices[aLevel+1][theVertexCount].color     = 0xFFFFFFFF;
                itsVertices[aLevel+1][theVertexCount].specular  = 0xFFFFFFFF;

                ComputeTextureCoordinates(  normal.x,
                                            normal.y,
                                            normal.z,
                                            &itsVertices[aLevel+1][theVertexCount]);

                theTable[u*itsNumVertices[aLevel] + v] = theVertexCount;
                theTable[v*itsNumVertices[aLevel] + u] = theVertexCount;

                theVertexCount++;
            }
        }
    assert(theVertexCount == itsNumVertices[aLevel+1]);

    //  Create the new faces.
    for (i = 0; i < itsNumFaces[aLevel]; i++)
    {
        //  The old vertices incident to this face will be f[0], f[1] and f[2].
        f = itsFaces[aLevel] + 3*i;

        //  The new ones will be ff[0], ff[1], and ff[2].
        //  Each ff[j] lies opposite the corresponding f[j].
        for (j = 0; j < 3; j++)
            ff[j] = theTable[f[(j+1)%3]*itsNumVertices[aLevel] + f[(j+2)%3]];

        //  Create the new faces.

        itsFaces[aLevel+1][12*i +  0] = ff[0];
        itsFaces[aLevel+1][12*i +  1] = ff[1];
        itsFaces[aLevel+1][12*i +  2] = ff[2];

        itsFaces[aLevel+1][12*i +  3] = f[0];
        itsFaces[aLevel+1][12*i +  4] = ff[2];
        itsFaces[aLevel+1][12*i +  5] = ff[1];

        itsFaces[aLevel+1][12*i +  6] = f[1];
        itsFaces[aLevel+1][12*i +  7] = ff[0];
        itsFaces[aLevel+1][12*i +  8] = ff[2];

        itsFaces[aLevel+1][12*i +  9] = f[2];
        itsFaces[aLevel+1][12*i + 10] = ff[1];
        itsFaces[aLevel+1][12*i + 11] = ff[0];
    }

    SAFE_DELETE_(theTable)
}
