//  SpinningEarth-GL.cpp
//
//       YOU DO NOT NEED TO READ THIS FILE TO UNDERSTAND
//             COMPUTER GRAPHICS IN CURVED SPACES.
//
//  CurvedSpaces-Simple includes the much simpler SceneryObjectCube
//  and SceneryObjectOctahedron, which illustrate the basic principles.
//
//  2002/09/24  Mirror-reverses the Earth, whose triangulation was
//  originally laid out in Direct3D's left-handed coordinate system,
//  to accomodate OpenGL's right-handed coordinate system.
//
//  2002/10/13  Replace icosahedron having no duplicated vertices
//  with an octahedron having duplicated vertices, to accomodate
//  the fact that OpenGL cannot wrap texture coordinates properly.
//  (That is, when OpenGL goes from (0.50, 0.99) to (0.50, 0.01)
//  is always backtracks "the long way" rather than wrapping around
//  "the short way".)
//
//  2003/01/10  Implement display lists to improve performance.

#include "SpinningEarth-GL.hpp"
#include <assert.h>
#include <math.h>

#define DEGREES_PER_SECOND  45
#define EARTH_RADIUS        0.05
#define PI                  3.14159265358979323846

//  To get the texture coordinates right on the duplicated vertices,
//  we temporarily separate them, one to either side of the branch cut.
#define TEMPORARY_OFFSET    0.001


//  The standard gluSphere() is easy to use,
//  but its rectangular triangulation scheme, aligned along
//  latitude and longitude lines, is woefully inefficient,
//  especially near the poles.  Use an efficient icosahedral
//  triangulation instead, subdivided as necessary.
//
//  2002/10/13  Icosahedron replaced with octahedron as
//  explained above.


CSpinningEarth::CSpinningEarth(void)
{
    int i;

    itsRotationAngle = 0.0;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        itsDisplayLists[i]          = 0;

        itsVertices[i]              = NULL;
        itsTextureCoordinates[i]    = NULL;
        itsFaces[i]                 = NULL;
        itsNumVertices[i]           = 0;
        itsNumFaces[i]              = 0;
    }

    InitOctahedron();

    for (i = 0; i < REFINEMENT_LEVEL; i++)
        SubdivideTriangulation(i);

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
        RemoveGap(i);
}

CSpinningEarth::~CSpinningEarth(void)
{
    int i;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        SAFE_DELETE_(itsVertices[i])
        SAFE_DELETE_(itsTextureCoordinates[i])
        SAFE_DELETE_(itsFaces[i])
    }

}


void CSpinningEarth::Update(double anElapsedTime)
{
    //  2002/09/24  Reverse the Earth's rotation to accomodate
    //  differing orientation conventions between Direct3D and OpenGL.
    itsRotationAngle -= DEGREES_PER_SECOND * anElapsedTime;

    if (itsRotationAngle < 0.0)
        itsRotationAngle += 360.0;
}


void CSpinningEarth::MakeDisplayLists(void)
{
    int i;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        itsDisplayLists[i] = glGenLists(1);
        if (itsDisplayLists[i] != 0)
        {
            glNewList(itsDisplayLists[i], GL_COMPILE);

                glEnableClientState(GL_VERTEX_ARRAY);
                glEnableClientState(GL_TEXTURE_COORD_ARRAY);

                glVertexPointer  (3, GL_DOUBLE, 0, itsVertices[i]          );
                glTexCoordPointer(2, GL_DOUBLE, 0, itsTextureCoordinates[i]);

//              If you have OpenGL 1.2, glDrawRangeElements() may be better than glDrawElements().
//              glDrawRangeElements(GL_TRIANGLES, 0, itsNumVertices[i] - 1, 3*itsNumFaces[i], GL_UNSIGNED_INT, itsFaces[i]);

                glDrawElements(GL_TRIANGLES, 3*itsNumFaces[i], GL_UNSIGNED_INT, itsFaces[i]);

                glDisableClientState(GL_VERTEX_ARRAY);
                glDisableClientState(GL_TEXTURE_COORD_ARRAY);

            glEndList();
        }
    }
}


void CSpinningEarth::FreeDisplayLists(void)
{
    int i;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        glDeleteLists(itsDisplayLists[i], 1);
        itsDisplayLists[i] = 0;
    }
}


void CSpinningEarth::RenderDisplayList(
    GLuint  aTexture,
    float   aDistance)
{
    int theRefinementLevel;

    glPushMatrix();
    glRotated(90, 1, 0, 0);
    glRotated(itsRotationAngle, 0, 0, 1);

    glBindTexture(GL_TEXTURE_2D, aTexture);

    //  Pick an appropriate level of detail at which
    //  to render this image of the spinning Earth.
    assert(REFINEMENT_LEVEL >= 3);
    if (aDistance < 1.0)
        theRefinementLevel = 3;
    else
    if (aDistance < 1.58)   //  just above pi/2
        theRefinementLevel = 2;
    else
    if (aDistance < 2.0)
        theRefinementLevel = 1;
    else
        theRefinementLevel = 0;

    glCallList(itsDisplayLists[theRefinementLevel]);

    glPopMatrix();
}


//  The remaining methods implement the octahedral triangulation
//  of the sphere.  Feel free to ignore them.

void CSpinningEarth::InitOctahedron(void)
{
    Vertex  v[7];
    int     i;
    GLuint  *f;

    itsNumVertices[0]           = 7;
    itsNumFaces[0]              = 8;
    itsVertices[0]              = new Vertex[7];
    itsTextureCoordinates[0]    = new TextureUV[7];
    itsFaces[0]                 = new GLuint[3*8];

    assert(itsVertices[0]           != NULL);
    assert(itsTextureCoordinates[0] != NULL);
    assert(itsFaces[0]              != NULL);

    v[0][0] =  0.0;
    v[0][1] =  0.0;
    v[0][2] =  1.0;

    v[1][0] =  1.0;
    v[1][1] =  +TEMPORARY_OFFSET;
    v[1][2] =  0.0;

    v[2][0] =  0.0;
    v[2][1] =  1.0;
    v[2][2] =  0.0;

    v[3][0] = -1.0;
    v[3][1] =  0.0;
    v[3][2] =  0.0;

    v[4][0] =  0.0;
    v[4][1] = -1.0;
    v[4][2] =  0.0;

    v[5][0] =  1.0;
    v[5][1] =  -TEMPORARY_OFFSET;
    v[5][2] =  0.0;

    v[6][0] =  0.0;
    v[6][1] =  0.0;
    v[6][2] = -1.0;

    for (i = 0; i < 7; i++)
    {
        itsVertices[0][i][0] = v[i][0] * EARTH_RADIUS;
        itsVertices[0][i][1] = v[i][1] * EARTH_RADIUS;
        itsVertices[0][i][2] = v[i][2] * EARTH_RADIUS;

//      itsNormals[0][i][0] = v[i][0];
//      itsNormals[0][i][1] = v[i][1];
//      itsNormals[0][i][2] = v[i][2];

        ComputeTextureCoordinates(v[i], itsTextureCoordinates[0][i]);
    }

    for (i = 0; i < 4; i++)
    {
        f = itsFaces[0] + i*6;

        f[0] = 0;
        f[1] = i+1;
        f[2] = i+2;

        f[3] = 6;
        f[4] = i+2;
        f[5] = i+1;
    }
}

void CSpinningEarth::SubdivideTriangulation(int aLevel)
{
    //  Subdivide the triangulation, replacing
    //  each old triangle with four new ones.
    //
    //    /\
    //   /__\
    //  /_\/_\
    //
    //  Write the subdivision to aLevel+1.

    unsigned    theNumExtraVertices;
    GLuint      u,
                v,
                *f,
                ff[3];
    double      normal[3],
                theLengthSquared,
                theFactor;
    int         *theTable,
                theVertexCount,
                i,
                j,
                k;

    assert( aLevel >=0
         && aLevel < REFINEMENT_LEVEL);

    //  The number of extra vertices needed at the cut
    //  will be 2^(aLevel + 1).
    theNumExtraVertices = (1 << (aLevel + 1));

    itsNumVertices[aLevel+1]        = itsNumVertices[aLevel] + 3*itsNumFaces[aLevel]/2 + theNumExtraVertices;
    itsNumFaces[aLevel+1]           = 4 * itsNumFaces[aLevel];
    itsVertices[aLevel+1]           = new Vertex[itsNumVertices[aLevel+1]];
    itsTextureCoordinates[aLevel+1] = new TextureUV[itsNumVertices[aLevel+1]];
    itsFaces[aLevel+1]              = new GLuint[3*itsNumFaces[aLevel+1]];

    assert(itsVertices[aLevel+1]           != NULL);
    assert(itsTextureCoordinates[aLevel+1] != NULL);
    assert(itsFaces[aLevel+1]              != NULL);

    //  Copy the vertices from the previous level.
    for (i = 0; i < itsNumVertices[aLevel]; i++)
    {
        for (j = 0; j < 3; j++)
            itsVertices[aLevel+1][i][j] = itsVertices[aLevel][i][j];
        for (j = 0; j < 2; j++)
            itsTextureCoordinates[aLevel+1][i][j] = itsTextureCoordinates[aLevel][i][j];
    }
    theVertexCount = itsNumVertices[aLevel];

    //  Create one new vertex on each edge.
    //  Make a table to index them, so the two triangles sharing
    //  an edge can share the same vertex.  (The memory for the
    //  table grows as the square of the number of old vertices.
    //  For modest triangulations this won't be a problem.
    //  For large triangulations a fancier algorithm could be used.)

    //  In spirit theTable is a 2-dimensional array.
    //  But such an array is a hassle to allocate in C/C++,
    //  so instead we just replace theTable[i][j] with
    //  theTable[i*itsNumVertices[aLevel] + j].
    theTable = new int[itsNumVertices[aLevel] * itsNumVertices[aLevel]];
    assert(theTable != NULL);

    for (i = 0; i < itsNumVertices[aLevel]; i++)
        for (j = 0; j < itsNumVertices[aLevel]; j++)
            theTable[i*itsNumVertices[aLevel] + j] = -1;

    for (i = 0; i < itsNumFaces[aLevel]; i++)
        for (j = 0; j < 3; j++)
        {
            u = itsFaces[aLevel][3*i +   j    ];
            v = itsFaces[aLevel][3*i + (j+1)%3];

            if (theTable[u*itsNumVertices[aLevel] + v] == -1)
            {
                assert(theVertexCount < itsNumVertices[aLevel+1]);

                theLengthSquared = 0.0;
                for (k = 0; k < 3; k++)
                {
                    normal[k] = (itsVertices[aLevel][u][k] + itsVertices[aLevel][v][k])/2;
                    theLengthSquared += normal[k]*normal[k];
                }
                theFactor = 1.0 / sqrt(theLengthSquared);
                for (k = 0; k < 3; k++)
                    normal[k] *= theFactor;
                //  This program doesn't use explicit lighting,
                //  but if desired you could record the normal vector
                //  at this point.
                
                for (k = 0; k < 3; k++)
                    itsVertices[aLevel+1][theVertexCount][k] = normal[k] * EARTH_RADIUS;

                ComputeTextureCoordinates(normal, itsTextureCoordinates[aLevel+1][theVertexCount]);

                theTable[u*itsNumVertices[aLevel] + v] = theVertexCount;
                theTable[v*itsNumVertices[aLevel] + u] = theVertexCount;

                theVertexCount++;
            }
        }
    assert(theVertexCount == itsNumVertices[aLevel+1]);

    //  Create the new faces.
    for (i = 0; i < itsNumFaces[aLevel]; i++)
    {
        //  The old vertices incident to this face will be f[0], f[1] and f[2].
        f = itsFaces[aLevel] + 3*i;

        //  The new ones will be ff[0], ff[1], and ff[2].
        //  Each ff[j] lies opposite the corresponding f[j].
        for (j = 0; j < 3; j++)
            ff[j] = theTable[f[(j+1)%3]*itsNumVertices[aLevel] + f[(j+2)%3]];

        //  Create the new faces.

        itsFaces[aLevel+1][12*i +  0] = ff[0];
        itsFaces[aLevel+1][12*i +  1] = ff[1];
        itsFaces[aLevel+1][12*i +  2] = ff[2];

        itsFaces[aLevel+1][12*i +  3] = f[0];
        itsFaces[aLevel+1][12*i +  4] = ff[2];
        itsFaces[aLevel+1][12*i +  5] = ff[1];

        itsFaces[aLevel+1][12*i +  6] = f[1];
        itsFaces[aLevel+1][12*i +  7] = ff[0];
        itsFaces[aLevel+1][12*i +  8] = ff[2];

        itsFaces[aLevel+1][12*i +  9] = f[2];
        itsFaces[aLevel+1][12*i + 10] = ff[1];
        itsFaces[aLevel+1][12*i + 11] = ff[0];
    }

    SAFE_DELETE_(theTable)
}

void CSpinningEarth::ComputeTextureCoordinates(
    Vertex      v,
    TextureUV   t)
{
    //  Technical note:  The north pole maps to texture
    //  coordinates (0.5, 1.0).  Thus half the arctic region
    //  is lost, and there's a texture discontinuity along
    //  the edge between two polar triangles.  This problem
    //  could be corrected by compressing a strip along
    //  the top edge of the bitmap from a rectangle to a
    //  triangle, but it's probably not worth the effort
    //  because (1) the discontinuity is small for a second order
    //  subdivision, and (2) the polar bitmap looks a bit
    //  pentagonal in any case.  Nevertheless, if in the future
    //  I want to fix the discontinuity, it can be done using
    //  DirectDrawSuface7::Blt().

    //  In theory atan2(0,0) returns 0, but who knows whether
    //  some future version of the math library might make
    //  trouble for us.
    if (v[0] != 0.0 || v[1] != 0.0)
    {
        t[0] = atan2(v[1], v[0])/(2*PI);
        if (t[0] < 0.0)
            t[0] += 1.0;
    }
    else
        t[0] = 0.5;

    t[1] = asin(v[2])/PI + 0.5;

    //  2002/09/24  Mirror-reverse the texture to accomodate
    //  differing orientation conventions between Direct3D and OpenGL.
    t[0] = 1.0 - t[0];
    
    //  For (unnecessary?) insurance against rounding errors...
    if (t[0] < 0.0) t[0] = 0.0;
    if (t[0] > 1.0) t[0] = 1.0;
    if (t[1] < 0.0) t[1] = 0.0;
    if (t[1] > 1.0) t[1] = 1.0;
}

void CSpinningEarth::RemoveGap(int aLevel)
{
    int i;

    for (i = 0; i < itsNumVertices[aLevel]; i++)
        if (fabs(itsVertices[aLevel][i][1]) <= TEMPORARY_OFFSET)
            itsVertices[aLevel][i][1] = 0.0;
}
